"""
Contains task to clear reduced dark and flat frames
"""

from __future__ import annotations

from tomoscan.scanbase import TomoScanBase as TomoscanScanBase

from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess

from tomwer.tasks.task import Task
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.scanutils import data_identifier_to_scan
from tomwer.core.reconstruction.darkflat import params as dkrf_reconsparams
from tomwer.utils import docstring


class ClearReducedDarkAndFlat(
    Task,
    SuperviseProcess,
    input_names=("data",),
    output_names=("data",),
):
    """
    Task to clear reduced darks and flats. Both on disk and on the object cache.
    th goal of this task is to make sure the scan is cleared of any reduced frames to reprocess it later.
    """

    def run(self):
        scan = self.inputs.data
        if not isinstance(scan, TomoscanScanBase):
            raise TypeError(
                f"scan should be an instance of {TomoscanScanBase}. Got {type(scan)}"
            )
        scan.set_reduced_flats(None)
        scan.reduced_flats_infos = None
        scan.set_reduced_darks(None)
        scan.reduced_darks_infos = None

        self.outputs.data = scan



# coding: utf-8
from __future__ import annotations

import logging
import os
from fnmatch import fnmatch

import h5py
from silx.gui import qt

from tomwer.core.process.control.datawatcher import status as datawatcherstatus
from tomwer.core.process.control.datawatcher.datawatcherobserver import (
    _DataWatcherObserver_MixIn,
    _DataWatcherStaticObserverMixIn,
    _OngoingObservation,
)
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory

logger = logging.getLogger(__name__)


class _QDataWatcherObserver(_DataWatcherObserver_MixIn, qt.QThread):
    """
    DatWatcherObserver implementation with a qt.QThread

    We have two implementations in order to avoid hard dependancy on qt for
    tomwer.core package
    """

    sigScanReady = qt.Signal(object)
    """emit when scan ready"""

    def __init__(
        self,
        obsMethod,
        observationClass,
        headDir=None,
        startByOldest=False,
        srcPattern=None,
        destPattern=None,
        ignoredFolders=None,
        file_name_pattern=None,
    ):
        qt.QThread.__init__(self)
        _DataWatcherObserver_MixIn.__init__(
            self,
            obsMethod=obsMethod,
            observationClass=observationClass,
            headDir=headDir,
            startByOldest=startByOldest,
            srcPattern=srcPattern,
            destPattern=destPattern,
            ignoredFolders=ignoredFolders,
            time_between_loops=0.2,
            file_name_pattern=file_name_pattern,
        )
        self.observations.sigScanReady.connect(self._signalScanReady)

    def _signalScanReady(self, scan):
        assert isinstance(scan, (TomwerScanBase, BlissScan))
        self.sigScanReady.emit(scan)

    def _getObserver(self, scanID):
        return _QDataWatcherStaticObserver(
            scanID=scanID,
            obsMethod=self.obsMethod,
            srcPattern=self.srcPattern,
            destPattern=self.destPattern,
            patternObs=self._patternObs,
            observationRegistry=self.observations,
        )

    def run(self):
        look_for_hdf5_file = self.obsMethod in (
            datawatcherstatus.BLISS_SCAN_END,
            datawatcherstatus.NXtomo_END,
        )

        def process(file_path):
            do_observation = (
                self.observations.isObserving(file_path) is False
                and (
                    (look_for_hdf5_file and h5py.is_hdf5(file_path))
                    or (self.dataWatcherProcess._isScanDirectory(file_path))
                )
                and file_path not in self.observations.ignoredFolders
            )
            if (
                do_observation
                and self.file_name_pattern is not None
                and not fnmatch(os.path.basename(file_path), self.file_name_pattern)
            ):
                do_observation = False
            if do_observation is True:
                # do the observation
                self.observe(file_path)

            if os.path.isdir(file_path):
                try:
                    for f in os.listdir(file_path):
                        full_file_path = os.path.join(file_path, f)
                        if os.path.isdir(full_file_path) or (
                            look_for_hdf5_file and h5py.is_hdf5(full_file_path)
                        ):
                            process(full_file_path)
                except Exception:
                    pass

        if not os.path.isdir(self.headDir):
            logger.warning("can't observe %s, not a directory" % self.headDir)
            return
        self.dataWatcherProcess = self._getDataWatcherProcess()
        process(self.headDir)

        self._processObservation()

    def waitForObservationFinished(self, timeOut=10):
        threads = list(self.observations.dict.values())
        for thread in threads:
            thread.wait(timeOut)

    def wait(self, *args, **kwargs):
        self.waitForObservationFinished()
        super(_QDataWatcherObserver, self).wait(*args, **kwargs)


class _QOngoingObservation(_OngoingObservation, qt.QObject):
    """
    _OngoingObservation with a QObject and signals for each event
    """

    sigScanReady = qt.Signal(object)
    """Emitted when a finished acquisition is detected"""
    sigObsAdded = qt.Signal(str)
    """Signal emitted when an observation is added"""
    sigObsRemoved = qt.Signal(str)
    """Signal emitted when an observation is removed"""
    sigObsStatusReceived = qt.Signal(str, str)
    """Signal emitted when receiving a new observation status"""

    def __init__(self):
        qt.QObject.__init__(self)
        _OngoingObservation.__init__(self)

    def _acquisition_ended(self, scanID):
        _OngoingObservation._acquisition_ended(self, scanID=scanID)
        try:
            scans = ScanFactory.create_scan_objects(
                scan_path=scanID, accept_bliss_scan=True
            )
        except Exception as e:
            logger.error(f"Fail to create TomoBase instance from {scanID} Error is {e}")
        else:
            for scan in scans:
                self.sigScanReady.emit(scan)

    def add(self, observer):
        already_observing = self.isObserving(observer.path)
        _OngoingObservation.add(self, observer=observer)
        if not already_observing:
            observer.sigStatusChanged.connect(self._updateStatus)
            self.sigObsAdded.emit(observer.path)

    def remove(self, observer):
        observing = self.isObserving(observer.path)
        _OngoingObservation.remove(self, observer=observer)
        if observing is True:
            observer.sigStatusChanged.disconnect(self._updateStatus)
            self.sigObsRemoved.emit(observer.path)

    def _updateStatus(self, status, scan):
        if self.isObserving(scan) is True:
            self.sigObsStatusReceived.emit(
                scan, datawatcherstatus.DICT_OBS_STATUS[status]
            )
        _OngoingObservation._updateStatus(self, status=status, scan=scan)

    def reset(self):
        # self.ignoreFolders = []
        for scanID, observer in self.dict:
            observer.sigStatusChanged.disconnect(self._updateStatus)
            observer.quit()
        self.dict = {}


class _QDataWatcherStaticObserver(_DataWatcherStaticObserverMixIn, qt.QThread):
    """
    Implementation of the _DataWatcherFixObserverMixIn with a qt.QThread
    """

    sigStatusChanged = qt.Signal(int, str)
    """signal emitted when the status for a specific directory change
    """

    def __init__(
        self,
        scanID,
        obsMethod,
        srcPattern,
        destPattern,
        patternObs,
        observationRegistry,
    ):
        qt.QThread.__init__(self)
        _DataWatcherStaticObserverMixIn.__init__(
            self,
            scanID=scanID,
            obsMethod=obsMethod,
            srcPattern=srcPattern,
            destPattern=destPattern,
            patternObs=patternObs,
            observationRegistry=observationRegistry,
        )

    def run(self):
        look_for_hdf5_file = self.obsMethod in (
            datawatcherstatus.BLISS_SCAN_END,
            datawatcherstatus.NXtomo_END,
        )
        if not look_for_hdf5_file and not os.path.isdir(self.path):
            logger.info("can't observe %s, not a directory" % self.path)
            self.status = "failure"
            self.sigStatusChanged.emit(
                datawatcherstatus.OBSERVATION_STATUS[self.status], self.path
            )
            self.validation = -1
            return

        try:
            scans = ScanFactory.create_scan_objects(
                scan_path=self.path, accept_bliss_scan=True
            )
        except ValueError as e:
            logger.error(e)
        else:
            for scan in scans:
                if scan.is_abort(
                    src_pattern=self.srcPattern, dest_pattern=self.destPattern
                ):
                    if self.status != "aborted":
                        logger.info("Acquisition %s has been aborted" % self.path)
                        self.dataWatcherProcess._removeAcquisition(
                            scanID=self.path, reason="acquisition aborted by the user"
                        )

                        self.status = "aborted"
                    self.sigStatusChanged.emit(
                        datawatcherstatus.OBSERVATION_STATUS[self.status], self.path
                    )
                    self.validation = -2
                    return
                dataComplete = self.dataWatcherProcess.is_data_complete()

                if dataComplete is True:
                    self.status = "acquisition ended"
                    self.sigStatusChanged.emit(
                        datawatcherstatus.OBSERVATION_STATUS[self.status], self.path
                    )
                    self.validation = 1
                else:
                    self.status = "waiting for acquisition ending"
                    self.sigStatusChanged.emit(
                        datawatcherstatus.OBSERVATION_STATUS[self.status], self.path
                    )
                    self.validation = 0
        return



# coding: utf-8
from __future__ import annotations


import os
import time

import pytest

from tomwer.core.process.control.datawatcher import DataWatcher
from tomwer.core.process.control.datawatcher.status import DET_END_XML, PARSE_INFO_FILE
from tomwer.core.utils.scanutils import MockEDF
from tomwer.core.utils.threads import LoopThread


@pytest.mark.skipif(os.name != "posix", reason="not tested under windows yet")
@pytest.mark.parametrize("det_method", (PARSE_INFO_FILE, DET_END_XML))
def test_data_watcher_io(tmp_path, det_method):
    """Test inputs and outputs types of the handler functions"""
    scan_folder = tmp_path / "folder" / "my_scan"
    data_watcher_process = DataWatcher()
    data_watcher_process.setWaitTimeBtwLoop(1)
    data_watcher_process.setObsMethod(det_method)

    MockEDF.mockScan(
        scanID=str(scan_folder), nRadio=10, nRecons=1, nPagRecons=4, dim=10
    )
    data_watcher_process.setFolderObserved(str(tmp_path / "folder"))
    data_watcher_process.set_serialize_output_data(True)
    LoopThread.quitEvent.clear()
    data_watcher_process.start()
    time.sleep(0.5)

    data_watcher_process.stop()
    data_watcher_process.waitForObservationFinished()
    import gc

    gc.collect()
    LoopThread.quitEvent.clear()



# coding: utf-8
from __future__ import annotations


import functools
import logging

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output
from processview.core import helpers as pv_helpers
from processview.core.manager import DatasetState, ProcessManager
from silx.gui import qt

import tomwer.core.process.reconstruction.sadeltabeta.sadeltabeta
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from orangecontrib.tomwer.orange.settings import CallbackSettingsHandler
from tomwer.core import settings
from tomwer.core.utils.lbsram import is_low_on_memory
from tomwer.core.cluster import SlurmClusterConfiguration
from tomwer.core.scan.scanbase import TomwerScanBase, _TomwerBaseDock
from tomwer.gui.reconstruction.sadeltabeta import (
    SADeltaBetaWindow as _SADeltaBetaWindow,
)
from tomwer.synctools.axis import QAxisRP
from tomwer.synctools.sadeltabeta import QSADeltaBetaParams
from tomwer.synctools.stacks.reconstruction.sadeltabeta import SADeltaBetaProcessStack

from ..utils import WidgetLongProcessing

_logger = logging.getLogger(__name__)


class SADeltaBetaWindow(_SADeltaBetaWindow):
    def __init__(self, Outputs, parent=None, process_id=None):
        _SADeltaBetaWindow.__init__(self, parent=parent)

        self.Outputs = Outputs
        self._sa_delta_beta_params = QSADeltaBetaParams()
        self._processing_stack = SADeltaBetaProcessStack(
            sa_delta_beta_params=self._sa_delta_beta_params, process_id=process_id
        )
        self._clusterConfig = None

    def setClusterConfig(self, cluster_config: dict):
        if not isinstance(
            cluster_config, (dict, type(None), SlurmClusterConfiguration)
        ):
            raise TypeError(
                f"cluster config is expected to be None, dict, {SlurmClusterConfiguration} not {type(cluster_config)}"
            )
        self._clusterConfig = cluster_config

    def _launchReconstructions(self):
        """callback when we want to launch the reconstruction of the
        slice for n cor value"""
        scan = self.getScan()
        if scan is None:
            return
        # step1: if isAutoFocus: validate automatically the scan
        # step2: update the interface if the current scan is the one displayed
        # else skip it
        callback = functools.partial(
            self._mightUpdateResult, scan, self.isAutoFocusLock()
        )
        self._processing_stack.add(
            data=scan, configuration=self.getConfiguration(), callback=callback
        )

    def _validate(self):
        self.validateCurrentScan()

    def _mightUpdateResult(self, scan: TomwerScanBase, validate: bool):
        if not isinstance(scan, TomwerScanBase):
            raise TypeError("scan is expected to be an instance of TomwerScanBase")
        if not isinstance(validate, bool):
            raise TypeError("validate is expected to be a boolean")
        if scan == self.getScan():
            self.setDBScores(
                scan.sa_delta_beta_params.scores,
                score_method=scan.sa_delta_beta_params.score_method,
            )
            if scan.sa_delta_beta_params.autofocus is not None:
                self.setCurrentDeltaBetaValue(scan.sa_delta_beta_params.autofocus)

            pm = ProcessManager()
            details = pm.get_dataset_details(
                dataset_id=scan.get_identifier(), process=self._processing_stack
            )
            current_state = ProcessManager().get_dataset_state(
                dataset_id=scan.get_identifier(), process=self._processing_stack
            )
            if current_state not in (
                DatasetState.CANCELLED,
                DatasetState.FAILED,
                DatasetState.SKIPPED,
            ):
                ProcessManager().notify_dataset_state(
                    dataset=scan,
                    process=self._processing_stack,
                    details=details,
                    state=DatasetState.WAIT_USER_VALIDATION,
                )
        if validate:
            self.validateScan(scan)

    def wait_processing(self, wait_time):
        self._processing_stack._computationThread.wait(wait_time)

    def validateCurrentScan(self):
        return self.validateScan(self.getScan())

    def validateScan(self, scan):
        if scan is None:
            return
        assert isinstance(scan, TomwerScanBase)
        selected_db_value = (
            self.getCurrentDeltaBetaValue() or scan.sa_delta_beta_params.autofocus
        )
        if selected_db_value is None:
            infos = f"no selected delta / beta value. {scan} skip SADeltaBetaParams"
            _logger.warning(infos)
            scan.sa_delta_beta_params.set_db_selected_value(None)
            pv_helpers.notify_skip(
                process=self._processing_stack, dataset=scan, details=infos
            )
        else:
            scan.sa_delta_beta_params.set_db_selected_value(selected_db_value)
            if scan.nabu_recons_params is not None:
                if "phase" not in scan.nabu_recons_params:
                    scan.nabu_recons_params["phase"] = {}
                scan.nabu_recons_params["phase"]["delta_beta"] = (selected_db_value,)
            _db_value = scan.sa_delta_beta_params.value
            infos = f"delta / beta selected for {scan}: {_db_value}"
            pv_helpers.notify_succeed(
                process=self._processing_stack, dataset=scan, details=infos
            )
        self.Outputs.data.send(scan)

    def getConfiguration(self) -> dict:
        config = super().getConfiguration()
        config["cluster_config"] = self._clusterConfig
        return config

    def setConfiguration(self, config: dict):
        # ignore slurm cluster. Defined by the upper widget
        config.pop("cluster_config", None)
        return super().setConfiguration(config)


class SADeltaBetaOW(SuperviseOW, WidgetLongProcessing):
    """
    Widget for semi-automatic delta / beta calculation

    behavior within a workflow:

    * no delta / beta value will be loaded even if an "axis" window exists on
      the upper stream.

    * if autofocus option is lock:

        * launch the series of reconstruction (with research width defined)
          and the estimated center of rotation if defined. Once the
          reconstruction is ended and if the autofocus button is still lock
          it will select the cor with the highest
          value and mode to workflow downstream.

    * hint: you can define a "multi-step" half-automatic center of rotation
      research by creating several "sa_delta_beta" widget and reducing the
      research width.

    Details about :ref:`sadeltabeta score calculation`
    """

    name = "multi-pag (sa-delta/beta calculation)"
    id = "orange.widgets.tomwer.sa_delta_beta"
    description = "Reconstruct a slice with several delta / beta values."
    icon = "icons/delta_beta_range.png"
    priority = 22
    keywords = [
        "multi",
        "multi-pag",
        "tomography",
        "semi automatic",
        "half automatic",
        "axis",
        "delta-beta",
        "delta/beta",
        "delta",
        "beta",
        "tomwer",
        "reconstruction",
        "position",
        "center of rotation",
        "sadeltabetaaxis",
        "sa_delta_beta_axis",
        "sa_delta_beta",
        "sadeltabeta",
    ]

    ewokstaskclass = (
        tomwer.core.process.reconstruction.sadeltabeta.sadeltabeta.SADeltaBetaTask
    )

    want_main_area = True
    resizing_enabled = True

    settingsHandler = CallbackSettingsHandler()

    sigScanReady = qt.Signal(TomwerScanBase)
    """Signal emitted when a scan is ready"""

    _ewoks_default_inputs = Setting({"data": None, "sa_delta_beta_params": None})

    class Inputs:
        data = Input(name="data", type=TomwerScanBase, default=True, multiple=False)
        data_recompute = Input(
            name="change recons params",
            type=_TomwerBaseDock,
            doc="recompute delta / beta",
        )
        cluster_in = Input(
            name="cluster_config",
            type=SlurmClusterConfiguration,
            doc="slurm cluster to be used",
            multiple=False,
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        """

        :param parent: QWidget parent or None
        """
        SuperviseOW.__init__(self, parent)
        WidgetLongProcessing.__init__(self)

        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._widget = SADeltaBetaWindow(
            Outputs=self.Outputs, parent=self, process_id=self.process_id
        )
        self._layout.addWidget(self._widget)

        sa_delta_beta_params = self._ewoks_default_inputs.get(
            "sa_delta_beta_params", None
        )
        self.setConfiguration(sa_delta_beta_params or {})

        # connect signal / slot
        self._widget.sigConfigurationChanged.connect(self._updateSettings)
        self._widget._processing_stack.sigComputationStarted.connect(
            self._startProcessing
        )
        self._widget._processing_stack.sigComputationEnded.connect(self._endProcessing)
        # expose API
        self.wait_processing = self._widget.wait_processing

    def __new__(cls, *args, **kwargs):
        # ensure backward compatibility with 'static_input'
        static_input = kwargs.get("stored_settings", {}).get("static_input", None)
        if static_input not in (None, {}):
            _logger.warning(
                "static_input has been deprecated. Will be replaced by _ewoks_default_inputs in the workflow file. Please save the workflow to apply modifications"
            )
            kwargs["stored_settings"]["_ewoks_default_inputs"] = static_input
        return super().__new__(cls, *args, **kwargs)

    def setConfiguration(self, configuration):
        if "workflow" in configuration:
            autofocus_lock = configuration["workflow"].get("autofocus_lock", None)
            if autofocus_lock is not None:
                self._widget.lockAutofocus(autofocus_lock)
            del configuration["workflow"]
        self._widget.setConfiguration(configuration)

    def getCurrentCorValue(self):
        return self._widget.getCurrentCorValue()

    def load_sinogram(self):
        self._widget.loadSinogram()

    def compute(self):
        self._widget.compute()

    def lockAutofocus(self, lock):
        self._widget.lockAutofocus(lock=lock)

    def isAutoFocusLock(self):
        return self._widget.isAutoFocusLock()

    @Inputs.data
    def process(self, scan):
        if scan is None:
            return
        if scan.axis_params is None:
            scan.axis_params = QAxisRP()
        if scan.sa_delta_beta_params is None:
            scan.sa_delta_beta_params = QSADeltaBetaParams()
        self._skipCurrentScan(new_scan=scan)

        if settings.isOnLbsram(scan) and is_low_on_memory(settings.get_lbsram_path()):
            self.notify_skip(
                scan=scan,
                details=f"sa-delta-beta has been skiped for {scan} because of low space in lbsram",
            )
            self.Outputs.data.send(scan)
        else:
            self._widget.setScan(scan=scan)
            self.notify_pending(scan)
            self.activateWindow()
            if self.isAutoFocusLock():
                self.compute()
            else:
                self.raise_()
                self.show()

    def _skipCurrentScan(self, new_scan):
        scan = self._widget.getScan()
        # if the same scan has been run several scan
        if scan is None or str(scan) == str(new_scan):
            return
        current_scan_state = ProcessManager().get_dataset_state(
            dataset_id=scan.get_identifier(), process=self
        )
        if current_scan_state in (
            DatasetState.PENDING,
            DatasetState.WAIT_USER_VALIDATION,
        ):
            details = "Was pending and has been replaced by another scan."
            self.notify_skip(scan=scan, details=details)
            self.Outputs.data.send(scan)

    @Inputs.data_recompute
    def reprocess(self, dataset):
        self.lockAutofocus(False)
        self.process(dataset)

    @Inputs.cluster_in
    def setCluster(self, cluster):
        self._widget.setClusterConfig(cluster_config=cluster)

    def validateCurrentScan(self):
        self._widget.validateCurrentScan()

    def _updateSettings(self):
        config = self._widget.getConfiguration()
        config.pop("cluster_config", None)
        self._ewoks_default_inputs = {
            "data": None,
            "sa_delta_beta_params": self._widget.getConfiguration(),
        }
        self._ewoks_default_inputs["sa_delta_beta_params"]["workflow"] = {
            "autofocus_lock": self._widget.isAutoFocusLock(),
        }

    def getConfiguration(self):
        return self._widget.getConfiguration()

    def cancel(self, scan):
        if scan is None:
            return
        if scan in self._widget._processing_stack:
            self._widget._processing_stack.remove(scan)
        if (
            self._widget._processing_stack._data_currently_computed.get_identifier().to_str()
            == scan.get_identifier().to_str()
        ):
            # stop current processing
            self._widget._processing_stack.cancel()
            # if possible process next
            if self._widget._processing_stack.can_process_next():
                self._widget._processing_stack._process_next()

    def getWaitingOverlay(self):
        return self._widget._tabWidget._resultsViewer._plot.getWaiterOverlay()

    def _startProcessing(self, *args, **kwargs):
        self.getWaitingOverlay().show()
        super()._startProcessing()

    def _endProcessing(self, *args, **kwargs):
        self.getWaitingOverlay().hide()
        super()._endProcessing()
